#include <iostream>
#include <sstream>
#include <fstream>
#include <iomanip>
#include <vector>
#include <string>
using namespace std;

enum FIELDS 
{PROGRAM=0, VERSION, TAPENR, FILE_CNTR, BLK_CNTR, DATE, MJD, UT_STRT, LST_STRT, COORD_TYPE,
 RA_STRT, DEC_STRT, GL_STRT, GB_STRT, FEED_ANGLE, OBS_TIME, COMMENT, NFILTER, CHANBW, NCHAN, FREQ_CHAN1,
 SAMP_INT, SAMP_GRP, GRP_BLK, BLK_SEC, FDCNTRL, DATA, UT, LST, RA, DEC, GL, GB, TEL_ZEN,
 TEL_AZ, ATTEN, TPOWER, NBLK_READ, SCAN_RATEL, SCAN_RATEB, NBEAM, IBEAM, PNAME, OBS_FILEN, NBITS, DDM, NDDCH, MOVE,
 PNTERR, TREE, NSYS, TELID, PANGLE, SMP_BSW, CAL_CYCLE, RESERVE};

const string fieldName[RESERVE+1] = 
{"PROGRAM", "VERSION", "TAPENR", "FILE_CNTR", "BLK_CNTR", "DATE", "MJD", "UT_STRT", "LST_STRT", "COORD_TYPE",
" RA_STRT", "DEC_STRT", "GL_STRT", "GB_STRT", "FEED_ANGLE", "OBS_TIME", "COMMENT", "NFILTER", "CHANBW", "NCHAN", "FREQ_CHAN1",
" SAMP_INT", "SAMP_GRP", "GRP_BLK", "BLK_SEC", "FDCNTRL", "DATA", "UT", "LST", "RA", "DEC", "GL", "GB", "TEL_ZEN",
" TEL_AZ", "ATTEN", "TPOWER", "NBLK_READ", "SCAN_RATEL", "SCAN_RATEB", "NBEAM", "IBEAM", "PNAME", "OBJS_FILEN", "NBITS", "DDM", "NDDCH", "MOVE",
" PNTERR", "TREE", "NSYS", "TELID", "PANGLE", "SMP_BSW", "CAL_CYCLE", "RESERVE"};

const int fieldDim[RESERVE+1]={1,1,1,1,1,1,1,1,1,1,
			 1,1,1,1,1,1,1,1,2,2,2,
			 2,2,1,1,1,2,1,1,1,1,1,1,1,
			 1,4,20,1,1,1,1,1,1,1,2,1,2,1,
			 1,1,1,1,1,1,1,1};

const int fieldSize[RESERVE+1] = {6,4,6,4,8,8,8,16,12,2,
			    16,16,8,8,8,8,64,2,8,4,12,
			    12,4,8,8,2,1,16,12,16,16,8,8,8,
			    8,4,4,8,8,8,4,4,16,16,2,8,4,2,
			    6,2,2,10,8,8,4,22};

int totalFieldSize[RESERVE+1];
int fieldStartLoc[RESERVE+1];

inline string getSub(string hdr, int field)
{return hdr.substr(fieldStartLoc[field],totalFieldSize[field]);};
			 

int main(int argc, char * argv[])
{
  /**
   *  This program simply reads a PM or PT tape data file block by block,
   *  and prints out the header information.
   *
   *  What it does is print out a list of all header parameters for
   *  the first block, giving the length in bytes of each of them, and
   *  then for each block in the file, prints out a select set of
   *  parameters.
   * 
   *  While it doesn't do anything with the data itself, it needs to
   *  know how big the data section is and be able to read through it
   *  correctly, otherwise the headers will be read incorrectly.
   */

  ifstream fin;
  string filename;
  if(argc==1){ 
//     filename = "../data/pttape_test";
    filename = "data/PM0001_001";
  }
  else if(argc==2){
    filename = argv[1];
  }
  else{
    cerr << "Either \"pm.x\" or \"pm.x filename\"...\n";
    return(1);
  }

  const int HDR_LENGTH=640;     // length of the header in bytes
  const int DAT_LENGTH=48*1024; // length of the data field in bytes

  for(int i=0;i<=RESERVE;i++) totalFieldSize[i] = fieldDim[i] * fieldSize[i];
  fieldStartLoc[0] = 0;
  for(int i=1;i<=RESERVE;i++) fieldStartLoc[i] = fieldStartLoc[i-1] + totalFieldSize[i-1];

  fin.open(filename.c_str(),ios::binary);
  if(!fin.good()){
    std::cerr << "Cannot find file " << filename << "!\nExiting...\n";
    exit(1);
  }
  // define and read in the header
  char *hdr = new char[4];
  fin.read( hdr, 4*sizeof(char) );
  hdr[3] = '\0';
  unsigned char compstring[4] = {'P', 'M', 'D', 0};
  if(strcmp((const char*)compstring,hdr)==0)
    std::cerr << "Yes!\n";
  else std::cerr << "No!\n";
  fin.close();

  fin.open(filename.c_str(),ios::binary);
  // define and read in the header
  hdr = new char[HDR_LENGTH];
  fin.read( hdr, HDR_LENGTH*sizeof(char) );
  string originalHDR(" ",HDR_LENGTH);
  for(int c=0;c<HDR_LENGTH;c++) originalHDR[c] = hdr[c];

  std::cerr << "Header field:\n" << hdr << std::endl;
  fin.close();

  std::cerr << "header length = " << originalHDR.size() << std::endl;

  std::cout << "Header parsing:\n";
  for(int i=0; i<=RESERVE; i++){
    std::cout << setw(10) << fieldName[i] << ":\t";
    std::cout << "[" << setw(3) << fieldStartLoc[i] << "]:\t";
    if(getSub(originalHDR,i)[0]!=0) std::cout << getSub(originalHDR,i);
    std::cout << "\n";
  }


  fin.open(filename.c_str(),ios::binary);
  // now try and read in the data
  char newone;
  
  string fullstring(" ",HDR_LENGTH), oldtime = "bob";
  while(!fin.read( hdr, HDR_LENGTH*sizeof(char) ).eof()){
//   for(int loop=0;loop<2;loop++){
//     fin.read( hdr, HDR_LENGTH*sizeof(char) );

    for(int c=0;c<HDR_LENGTH;c++) fullstring[c] = hdr[c];

    float blk_sec = atof(getSub(fullstring,BLK_SEC).c_str());
    int grp_blk = atoi( getSub(fullstring,GRP_BLK).c_str() );
    stringstream ss;
    ss.str(getSub(fullstring,NCHAN));
    int nch[2];
    ss >> nch[0] >> nch[1];

    std::cerr << setw(totalFieldSize[BLK_CNTR]) << getSub(fullstring,BLK_CNTR)
	      << " "
	      << setw(totalFieldSize[DATE])     << getSub(fullstring,DATE)
	      << setw(totalFieldSize[UT_STRT])  << getSub(fullstring,UT_STRT)
	      << setw(totalFieldSize[UT])       << getSub(fullstring,UT)
	      << setw(totalFieldSize[RA_STRT])  << getSub(fullstring,RA_STRT)
	      << setw(totalFieldSize[DEC_STRT]) << getSub(fullstring,DEC_STRT)
	      << setw(totalFieldSize[RA])       << getSub(fullstring,RA)
	      << setw(totalFieldSize[DEC])      << getSub(fullstring,DEC)
	      << setw(totalFieldSize[MOVE])     << getSub(fullstring,MOVE)
	      << setw(totalFieldSize[IBEAM])    << getSub(fullstring,IBEAM)
// 	      << " " << blk_sec / float(grp_blk) 
// 	      << " " << 48.*1024.*8. / (nch[0] + nch[1])
// 	      << " " << 48.*1024.*8. / (nch[0] + nch[1]) * blk_sec / float(grp_blk) 
	      << " " << setw(totalFieldSize[MOVE]) << getSub(fullstring,MOVE)
	      << std::endl;
  
    for(int i=0;i<DAT_LENGTH;i++) fin.read( &newone, sizeof(char) ); 

   }


}

